package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.box.Box
import com.ionspin.kotlin.crypto.box.BoxCorruptedOrTamperedDataException
import com.ionspin.kotlin.crypto.box.crypto_box_NONCEBYTES
import com.ionspin.kotlin.crypto.scalarmult.ScalarMultiplication
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.crypto2.Asymmetric.Message
import net.sergeych.crypto2.Asymmetric.PublicKey
import net.sergeych.crypto2.Asymmetric.SecretKey

/**
 * Public-key encryption implementation. Generally should be libsodium-compatible.
 *
 * ## How to
 *
 * - [Asymmetric.generateKeys] create a key pair. Keys are serializable.
 * - [SecretKey] provides authenticated encryption for a [PublicKey] receiver.
 * - [PublicKey] provides decryption and anonymous encryption.
 * - [Message] is a serializable container with encrypted message and all necessary data to decrypt it.
 *
 * __Algorithms:__
 *
 * - Key exchange: X25519.
 * - Encryption: XSalsa20
 * - Authentication: Poly1305
 */
object Asymmetric {

    /**
     * Encrypted message holder.
     *
     * Do not instantiate it directly, use [SecretKey.encrypt], [PublicKey.encryptAnonymously] or [createMessage]
     * instead. Also [SecretKey.decrypt] can be used to decrypt it same as [decrypt] or [decryptWithSenderKey].
     *
     * To successfully decrypt the message, it is necessary to know a sender public key, and non-secret nonce.
     * This class carries all this information; serialize and pass it to the recipient.
     */
    @Serializable
    class Message(
        private val nonce: UByteArray,
        private val encryptedMessage: UByteArray,
        val senderPublicKey: PublicKey? = null,
    ) {
        /**
         * Decrypt the message, same as [SecretKey.decrypt]
         */
        fun decrypt(recipientKey: SecretKey): UByteArray {
            check(senderPublicKey != null)
            return decryptWithSenderKey(senderPublicKey, recipientKey)
        }

        /**
         * Decrypt a message which is not include sender's public key (which should somehow be
         * known to the recipient). Use it if [senderPublicKey] is null.
         */
        fun decryptWithSenderKey(senderKey: PublicKey, recipientKey: SecretKey): UByteArray {
            return try {
                Box.openEasy(encryptedMessage, nonce, senderKey.keyBytes, recipientKey.keyBytes)
            } catch (_: BoxCorruptedOrTamperedDataException) {
                throw DecryptionFailedException()
            }
        }
    }

    /**
     * Encrypt the [plainData] using [from] sender for [recipient] public key. Note that to decrypt it
     * the [SecretKey] that corresponds to the [recipient] public key is needed, Sender can't decrypt the message!
     *
     * The authenticated encryption is used, is the message _is successfully decrypted_, it also means that
     * it was signed by the sender, whose public key is known at the decryption time.
     *
     * When it is important not to provide senders' key, use [PublicKey.encryptAnonymously].
     *
     * @param from the senders' secret key.
     * @param recipient the recipients' public key.
     * @param plainData data to encrypt
     * @param excludeSenderKey if set to true, senders' public key will not be included in the message.
     *      In this case, the recipient must know it from other sources to be able to decrypt the message.
     */
    fun createMessage(
        from: SecretKey, recipient: PublicKey, plainData: UByteArray,
        excludeSenderKey: Boolean = false,
    ): Message {
        val nonce = randomNonce()
        return Message(
            nonce,
            Box.easy(plainData, nonce, recipient.keyBytes, from.keyBytes),
            if (excludeSenderKey) null else from.publicKey
        )
    }

    /**
     * The generated key pair. See [generateKeys]
     */
    data class KeyPair(val secretKey: SecretKey,val senderKey: PublicKey)

    /**
     * Generate a new random pair of public and secret keys.
     */
    fun generateKeys(): KeyPair {
        val p = Box.keypair()
        val pk = PublicKey(p.publicKey)
        return KeyPair(SecretKey(p.secretKey, pk), pk)
    }

    private fun randomNonce(): UByteArray = randomBytes(crypto_box_NONCEBYTES)

    /**
     * The public key used as the recipient for [Message] (see [SecretKey.encrypt], etc.). It also
     * could be used to encrypt anonymous messages with [encryptAnonymously]
     */
    @Serializable
    class PublicKey(val keyBytes: UByteArray) {
        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is PublicKey) return false

            return keyBytes contentEquals other.keyBytes
        }

        /**
         * Create an anonymous message that could be decrypted only with the [SecretKey] that corresponds this.
         * Anonymous message uses one-time secret key, the public part of which is included into the
         * [Message], so the sender could not be identified.
         * The authentication is used despite the anonymity, and the fact of the successful decryption
         * proves that the message was not altered after creation.
         */
        fun encryptAnonymously(plainData: UByteArray): Message =
            createMessage(generateKeys().secretKey, this, plainData)

        override fun hashCode(): Int {
            return keyBytes.hashCode()
        }
    }

    /**
     * The secret key
     */
    @Serializable
    class SecretKey(
        val keyBytes: UByteArray,
        @Transient
        val _cachedPublicKey: PublicKey? = null,
    ) {

        /**
         * Encrypt the message for [recipient]. The [Message] will include the our [publicKey]
         * and thus ready to be decrypted.
         *
         * The message is authenticated by this secret key.
         */
        fun encrypt(plainData: UByteArray,recipient: PublicKey): Message =
            createMessage(this, recipient, plainData)

        /**
         * Decrypt with authentication checks the message which must have [Message.senderPublicKey] set.
         * Use [decryptWithSenderKey] otherwise. Note that the authenticated encryption is always use, even if
         * the [PublicKey.encryptAnonymously] was used to create a message, if it is successfully decrypted,
         * it is guaranteed that the message was not altered after creation.
         *
         * @throws DecryptionFailedException If the message is tampered (changed after creation) or was not intended for us,
         */
        fun decrypt(message: Message): UByteArray = message.decrypt(this)


        /**
         * Decrypt using [senderPublicKey] as a sender key (overriding the [Message.senderPublicKey] if set).
         * See [decrypt] for more.
         */
        fun decryptWithSenderKey(message: Message,senderPublicKey: PublicKey): UByteArray =
            message.decryptWithSenderKey(senderPublicKey,this)

        private var cachedPublicKey: PublicKey? = _cachedPublicKey

        /**
         * The corresponding public key
         */
        val publicKey: PublicKey by lazy {
            PublicKey(ScalarMultiplication.scalarMultiplicationBase(keyBytes))
                .also { cachedPublicKey = it }
        }

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is SecretKey) return false

            return keyBytes contentEquals other.keyBytes
        }

        override fun hashCode(): Int {
            return keyBytes.hashCode()
        }
    }

}